#!/usr/local/bin/perl

#############################################################################
#
# File:        flusher
# Who:         Jan Sundberg, dKraft AB
#              sundberg@dkraft.se
#              070 - 595 80 86
# Description: A program that repeatedly reads input up to a specified 
#              buffer size and appends the same output to either file 
#              in an open/close scheme. Suitable for producing 
#              log files for on-the-fly truncation.
#              Use -l option for using lines instead of bytes
#
#############################################################################

use Getopt::Std;

$usage = "Usage: flusher [-l] [-b bufsize[k|m]] [file]\n";

# get buffer size option
getopts ('lb:') || exit print $usage;

# default buffer size is 4 kB
$opt_b ||= "4k";

# adjust for kB or MB sizes
$opt_b = $1 * 1024 if $opt_b =~ /(\d+)k$/i;
$opt_b = $1 * 1024 ** 2 if $opt_b =~ /(\d+)m$/i;

# $opt_b now has to be numeric
exit print $usage if $opt_b =~ /\D/;

# get output file, if any
$file = shift @ARGV;
exit print $usage if @ARGV;

do {
    local $buf = undef;
    if ($opt_l) {
	# read $opt_b lines into $buf
	local $lines = $opt_b;
	$buf .= <STDIN> while $lines-- && ! eof STDIN;
    } else {
	# read exactly $opt_b bytes into $buf
	for ($offset = undef;
	     $read = read STDIN, $buf, $opt_b - $offset, $offset;
	     $offset += $read) { }
    }

    # squirt it out to the file specified
    if ($file) {
        open OUT, ">>$file";
        print OUT $buf;
        close OUT;
    } else {
        print STDOUT $buf;
    }
} until eof STDIN;
