#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "io.h"
#include "cr.h"
#include "logparse.h"
#include "db.h"
#include "global.h"

void db_AddPlayer(Player *p)
/* This fuction allocates memory and adds a Player to the hash-table
	I might be out on thin ice on how to manage those pesky pointers/arrays
	This is only used when we are reading a database. The dynamic adding
	routine is FindElseAddPlayer, since you always want a Player-struct
	returned while parsing.
*/

{
	Playerlist *inlist;

/* ALERT! This way of looking up the index could be totally wrong.
	Consider declaring the function with Playerlist *list[]. */	

	inlist = plist[lp_HashFunction(p->name)];
	if (inlist != NULL) {
		while (inlist->next != NULL)
			inlist = inlist->next;
		inlist->next = calloc(1, sizeof(Playerlist));
		inlist = inlist->next;
	} else {
		inlist = (plist[lp_HashFunction(p->name)] = calloc(1, sizeof(Playerlist)));
	}
	inlist->p = p;
	inlist->next = NULL;	

}

Player *db_FindElseAddPlayer(char *name)
/* Search out the player with the name name, by first converting
	it to visible ascii and then looking up the hashtable. If it's
	not there, create the entry and return that zeroed Player. */
{
	Playerlist	*searchlist, *prev=NULL;
	Player		*p=NULL;
	char			*pname;
	char			tmpname[NAME_LENGTH];
	
	strcpy(tmpname, name);			/* Don't want to work on constant string */

	lp_FixunixdedName(tmpname);
	lp_MatchName(tmpname);			/* Correct name to normal ascii */
/*	printf(" Finally: %s\n", tmpname);fflush(stdout);*/

/* ALERT! This way of checking out index could be dead wrong.
	Consider declaring the header with Playerlist *list[] instead.
*/
	searchlist = plist[lp_HashFunction(tmpname)]; /* Find its location in the table */

	pname = calloc(1, NAME_LENGTH);

/*	while (searchlist != NULL && strcmp(tmpname,pname)!=0) {
		searchlist = ((Playerlist *)searchlist)->next;
		p = ((Playerlist *)searchlist)->p;
		pname = ((Player *)p)->name;
	}
*/
		
	while (searchlist != NULL && strcmp(tmpname, pname)!=0) {
		p = searchlist->p;
		if (p != NULL)
			strcpy(pname, p->name);
		prev=searchlist;
		searchlist = searchlist->next;
	}
	
	if (strcmp(tmpname,pname)!=0)
	{
		globs.nbrplayers++;	/* Increase the total number of players */
		if (prev!=NULL) {
			prev->next = calloc(1, sizeof(Playerlist));
			searchlist = prev->next;
		} else {
			plist[lp_HashFunction(tmpname)] = calloc(1, sizeof(Playerlist));
			searchlist = plist[lp_HashFunction(tmpname)];
		}
			searchlist->p = calloc(1,sizeof(Player));
			searchlist->next = NULL;
			p = searchlist->p;
		strcpy(p->name, tmpname);
	}
	
	free(pname);
	return p;
}

float db_KillsMin(Player *p)
{
	return (p->time_on==0?0:60.0*lp_GetKills(p)/p->time_on);
}

float db_DeathsMin(Player *p)
{
	return (p->time_on==0?0:60.0*lp_GetDeaths(p)/p->time_on);
}

float db_ScoreMin(Player *p)
{
	return(p->time_on==0?0:60.0*lp_GetScore(p)/p->time_on);
}

void db_qsort(int lo, int up, int n)
{
	int		i, j;
	Player	*p;
	while (up > lo) {
		i = lo;
		j = up;
		p = table[lo];
		while ( i<j ) {
			for ( ; db_is_largest(table[j], p) && j>0; j--);
			for ( table[i] = table[j]; i<j && (!db_is_largest(table[i], p)); i++);
			table[j] = table[i];
		}	
		table[i] = p;
		if ( i-lo < up-i ) {
			db_qsort(lo, i-1, n);
			lo = i+1;
		} else {
			db_qsort(i+1, up, n);
			up = i-1;
		}
	}
}

int db_is_largest(Player *p1, Player *p2)
{      
   switch (inits.sortitem) {
   case SI_FRANK			: return (lp_GetScore(p1)<lp_GetScore(p2)); break;
   case SI_KILLS			: return (lp_GetKills(p1)<lp_GetKills(p2)); break;
   case SI_DEATHS			: return (lp_GetDeaths(p1)<lp_GetDeaths(p2)); break;
   case SI_SUICIDES		: return (p1->suicides<p2->suicides); break;
   case SI_CAPTURES		: return (p1->captures<p2->captures); break;
   case SI_DROPS			: return (p1->drops<p2->drops); break;
   case SI_PICKUPS		: return (p1->pickups<p2->pickups); break;
   case SI_RETURNS		: return (p1->returns<p2->returns); break;
   case SI_K_AXE			: return (p1->kills[AXE]<p2->kills[AXE]); break;
   case SI_K_BOOMSTICK	: return (p1->kills[BOOMSTICK]<p2->kills[BOOMSTICK]); break;
   case SI_K_SUPER_SHOTGUN	: return (p1->kills[SUPER_SHOTGUN]<p2->kills[SUPER_SHOTGUN]); break;
	case SI_K_NAILGUN		: return (p1->kills[NAILGUN]<p2->kills[NAILGUN]); break;
   case SI_K_SUPER_NAILGUN	: return (p1->kills[SUPER_NAILGUN]<p2->kills[SUPER_NAILGUN]); break;
   case SI_K_GRENADE_LAUNCHER	: return (p1->kills[GRENADE_LAUNCHER]<p2->kills[GRENADE_LAUNCHER]); break;
	case SI_K_ROCKET_LAUNCHER	: return (p1->kills[ROCKET_LAUNCHER]<p2->kills[ROCKET_LAUNCHER]); break;
	case SI_K_THUNDERBOLT	: return (p1->kills[THUNDERBOLT]<p2->kills[THUNDERBOLT]); break;   
	case SI_K_GRAPPLING_HOOK	: return (p1->kills[GRAPPLING_HOOK]<p2->kills[GRAPPLING_HOOK]); break;
	case SI_K_TELEFRAG    : return (p1->kills[TELEFRAG]<p2->kills[TELEFRAG]); break;	
   case SI_D_AXE			: return (p1->deaths[AXE]<p2->deaths[AXE]); break;
   case SI_D_BOOMSTICK	: return (p1->deaths[BOOMSTICK]<p2->deaths[BOOMSTICK]); break;
   case SI_D_SUPER_SHOTGUN	: return (p1->deaths[SUPER_SHOTGUN]<p2->deaths[SUPER_SHOTGUN]); break;
	case SI_D_NAILGUN		: return (p1->deaths[NAILGUN]<p2->deaths[NAILGUN]); break;
   case SI_D_SUPER_NAILGUN	: return (p1->deaths[SUPER_NAILGUN]<p2->deaths[SUPER_NAILGUN]); break;
   case SI_D_GRENADE_LAUNCHER	: return (p1->deaths[GRENADE_LAUNCHER]<p2->deaths[GRENADE_LAUNCHER]); break;
	case SI_D_ROCKET_LAUNCHER	: return (p1->deaths[ROCKET_LAUNCHER]<p2->deaths[ROCKET_LAUNCHER]); break;
	case SI_D_THUNDERBOLT	: return (p1->deaths[THUNDERBOLT]<p2->deaths[THUNDERBOLT]); break;   
	case SI_D_GRAPPLING_HOOK	: return (p1->deaths[GRAPPLING_HOOK]<p2->deaths[GRAPPLING_HOOK]); break;
	case SI_D_TELEFRAG	: return (p1->deaths[TELEFRAG]<p2->deaths[TELEFRAG]); break;
	case SI_CONNECTS		: return (p1->connects<p2->connects); break;
	case SI_TIMEON			: return (p1->time_on<p2->time_on); break;
   case SI_QRANK			: return (lp_Qrank(p1)<lp_Qrank(p2)); break;
	case SI_KILLSMIN		: return (db_KillsMin(p1)<db_KillsMin(p2));
	case SI_DEATHSMIN		: return (db_DeathsMin(p1)<db_DeathsMin(p2));
   case SI_KP_AXE			: return (lp_GetKPerc(p1, AXE)<lp_GetKPerc(p2, AXE)); break;
   case SI_KP_BOOMSTICK	: return (lp_GetKPerc(p1, BOOMSTICK)<lp_GetKPerc(p2, BOOMSTICK)); break;
   case SI_KP_SUPER_SHOTGUN	: return (lp_GetKPerc(p1, SUPER_SHOTGUN)<lp_GetKPerc(p2, SUPER_SHOTGUN)); break;
	case SI_KP_NAILGUN		: return (lp_GetKPerc(p1, NAILGUN)<lp_GetKPerc(p2, NAILGUN)); break;
   case SI_KP_SUPER_NAILGUN	: return (lp_GetKPerc(p1, SUPER_NAILGUN)<lp_GetKPerc(p2, SUPER_NAILGUN)); break;
   case SI_KP_GRENADE_LAUNCHER	: return (lp_GetKPerc(p1, GRENADE_LAUNCHER)<lp_GetKPerc(p2, GRENADE_LAUNCHER)); break;
	case SI_KP_ROCKET_LAUNCHER	: return (lp_GetKPerc(p1, ROCKET_LAUNCHER)<lp_GetKPerc(p2, ROCKET_LAUNCHER)); break;
	case SI_KP_THUNDERBOLT	: return (lp_GetKPerc(p1, THUNDERBOLT)<lp_GetKPerc(p2, THUNDERBOLT)); break;   
	case SI_KP_GRAPPLING_HOOK	: return (lp_GetKPerc(p1, GRAPPLING_HOOK)<lp_GetKPerc(p2, GRAPPLING_HOOK)); break;
	case SI_KP_TELEFRAG    : return (lp_GetKPerc(p1, TELEFRAG)<lp_GetKPerc(p2, TELEFRAG)); break;
   case SI_DP_AXE			: return (lp_GetDPerc(p1, AXE)<lp_GetDPerc(p2, AXE)); break;
   case SI_DP_BOOMSTICK	: return (lp_GetDPerc(p1, BOOMSTICK)<lp_GetDPerc(p2, BOOMSTICK)); break;
   case SI_DP_SUPER_SHOTGUN	: return (lp_GetDPerc(p1, SUPER_SHOTGUN)<lp_GetDPerc(p2, SUPER_SHOTGUN)); break;
	case SI_DP_NAILGUN		: return (lp_GetDPerc(p1, NAILGUN)<lp_GetDPerc(p2, NAILGUN)); break;
   case SI_DP_SUPER_NAILGUN	: return (lp_GetDPerc(p1, SUPER_NAILGUN)<lp_GetDPerc(p2, SUPER_NAILGUN)); break;
   case SI_DP_GRENADE_LAUNCHER	: return (lp_GetDPerc(p1, GRENADE_LAUNCHER)<lp_GetDPerc(p2, GRENADE_LAUNCHER)); break;
	case SI_DP_ROCKET_LAUNCHER	: return (lp_GetDPerc(p1, ROCKET_LAUNCHER)<lp_GetDPerc(p2, ROCKET_LAUNCHER)); break;
	case SI_DP_THUNDERBOLT	: return (lp_GetDPerc(p1, THUNDERBOLT)<lp_GetDPerc(p2, THUNDERBOLT)); break;   
	case SI_DP_GRAPPLING_HOOK	: return (lp_GetDPerc(p1, GRAPPLING_HOOK)<lp_GetDPerc(p2, GRAPPLING_HOOK)); break;
	case SI_DP_TELEFRAG    : return (lp_GetDPerc(p1, TELEFRAG)<lp_GetDPerc(p2, TELEFRAG)); break;
	case SI_VICTORIES		: return (p1->victories<p2->victories); break;
	case SI_DEFEATS		: return (p1->defeats<p2->defeats); break;
	case SI_TIES			: return (p1->ties<p2->ties); break;
	case SI_CHICKENS		: return (p1->chickens<p2->chickens); break;
	case SI_BRAVES			: return (p1->braves<p2->braves); break;
	case SI_CAPRATIO		: return (lp_CapRatio(p1)<lp_CapRatio(p2)); break;

   default					: io_Error("No sorttype specified\n"); return 0; break;
	}

}        


int db_MakeTable()
/* Returns the number of entries left after filering away minkills. */
{
	int			i, tc=0;
	Playerlist	*pl;
	
/*	Allocate memory for an array of players. */
/*	if ((table = calloc(players, sizeof(Player *)))==NULL)
		io_Error("Memory exhausted in db_MakeTable.\n"); */

/* Get all players and put them in the unsorted table. */	
	for (i=0;i<HASH_TABLE_SIZE;i++) {
		pl = plist[i];
		while (pl != NULL) {
		/* Filter away minkills */
			if (lp_GetKills(pl->p)>=inits.minkills)
				table[tc++] = pl->p;
			pl = pl->next;
		}

	}
	
	/* Use quick-sort to sort them */
	
	#ifdef TABLE_DEBUG
	printf("%d player structs read, expected %d.\n", tc, players);
	#endif TABLE_DEBUG
	db_qsort(0, tc-1, tc-1); 
	
	/* Reverse table if SO_WORST */
	if (inits.sortorder == SO_WORST) {
		int		rev=0;
		Player 	*prev;
		while (rev<((tc-1)/2))
			{
				prev = table[rev];
				table[rev] = table[tc-rev-1];
				table[tc-(rev++)-1] = prev;
			}
	}
	
	return tc; 
}

void db_Free()
/*	Give the system back all memory allocated by the players. */
{
	int			i;
	Playerlist	*pl, *pl2;
	
	for (i=0;i<HASH_TABLE_SIZE;i++)
	{
		pl = plist[i];
		while(pl!=NULL)
		{
			pl2 = pl->next;
			free(pl->p);
			free(pl);
			pl = pl2;
		}
	}
}

void db_ReadDBase()
/*	Reads the players in a database file.
	Return value is the number of players in the file.
*/
{
	FILE		*dbf;
	Player	*p;
	char		*str;
	float		ver;
	
	if ((dbf = fopen(files.dbfile, "r")) == NULL)
		io_Error("Couldn\'t open existing database for reading.\n");
	
	if ((str = calloc(1, 256))==NULL)	 /* Don't forget to dealloc this memory */
		io_Error("Memory exhausted in db_ReadDBase().\n");
	
	fscanf(dbf, "%[^\n]\n", str);
	
/*	Check if we have a ctfstat database header in the specified file */
	if (strncmp(str, HEADER, 2)!=0) 
		io_Error("Invalid database header.\n");
/*	Compare the version number. I don't use io_Error here. Lousy.. */
	str += 2;
	ver = 1.0 * atoi(str) / 10;
	str -= 2;
	if (abs(ver - CR_VERSION)>0.001 ) 
		io_Error("This v%.2f database isn't compatible with ctfstat v%.2f\n", ver, CR_VERSION);
/*	Finished checking. Now go on with the reading of entries */
/*	First read the current score for the red and blue team */

	fscanf(dbf, "%d %d %d\n", &globs.red_score, &globs.blue_score, &globs.flagreg);

/*	Continue with reading the entries */
	
	if ((p = calloc(1, sizeof(Player)))== NULL)
		io_Error("Memory exhausted while reading entries from db.\n");

	while (fscanf(dbf, "%[^\n]\n",p->name)==1) {
		
		if (fscanf(dbf, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %f %f %f %d %d %d %d %d %d %d\n",
		&p->kills[0], &p->kills[1], &p->kills[2], &p->kills[3], &p->kills[4], &p->kills[5], &p->kills[6], &p->kills[7], &p->kills[8], &p->kills[9],
		&p->deaths[0], &p->deaths[1], &p->deaths[2], &p->deaths[3], &p->deaths[4], &p->deaths[5], &p->deaths[6], &p->deaths[7], &p->deaths[8], &p->deaths[9],
		&p->captures, &p->drops, &p->pickups, &p->returns, &p->suicides,
		&p->kill_score, &p->death_score, 
		&p->time_on, &p->connects, &p->current_team,&p->victories,&p->defeats,&p->ties,
		&p->chickens,&p->braves)!=35)
			io_Error("Corrupt database.\n");

		db_AddPlayer(p);
		
		if ((p = calloc(1, sizeof(Player)))== NULL)
			io_Error("Memory exhausted while reading entries in db_ReadDBase().\n");
		globs.nbrplayers++;
	}
	
	free(str);
}

void db_WriteDBase()
{
	FILE			*dbf;
	Player		*p;
	Playerlist	*pl;
	int			i;
	
		if ((dbf = fopen(files.dbfile, "w")) == NULL)
			io_Error("Couldn\'t open new database for writing.\n");

		/* Write the header and the current team-scores */
		fprintf(dbf, "%s\n", HEADER);
		fprintf(dbf, "%d %d %d\n", globs.red_score, globs.blue_score, globs.flagreg);
		
		for (i=0; i<HASH_TABLE_SIZE; i++)
		{
			pl = plist[i];
			while (pl != NULL) {
				p = pl->p;
				if (strcmp(p->name, "")!=0) {
					fprintf(dbf, "%s\n", p->name);
					fprintf(dbf, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %f %f %f %d %d %d %d %d %d %d\n",
					p->kills[0], p->kills[1], p->kills[2], p->kills[3], p->kills[4], p->kills[5], p->kills[6], p->kills[7], p->kills[8], p->kills[9],
					p->deaths[0], p->deaths[1], p->deaths[2], p->deaths[3], p->deaths[4], p->deaths[5], p->deaths[6], p->deaths[7], p->deaths[8], p->deaths[9],
					p->captures, p->drops, p->pickups, p->returns, p->suicides,
					p->kill_score, p->death_score,
					p->time_on, p->connects, p->current_team, p->victories, p->defeats, p->ties,
					p->chickens, p->braves);
				}
				
				pl = pl->next;
				
			}			
		}
}

