#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "io.h"
#include "cr.h"
#include "logparse.h"
#include "db.h"
#include "html.h"
#include "global.h"

float lp_CapRatio(Player *p)
/*	Calculates the capture ratio */
{
	if ( (p->captures + p->drops) !=0 )
		return (float) p->captures / (p->captures + p->drops);
	else
		return 0;
}

int lp_HashFunction(char *name)
/* A somewhat coarse hash-function, which I may improve later. */
{
	unsigned int	hashed=1;
	int				i;

	for (i=0;i<strlen(name) && i<HASH_GRANULARITY;i++)
	{
		#ifdef HASH_DEBUG
		printf("%d:", name[i]);
		#endif
		hashed *=name[i];
	}
	
	#ifdef HASH_DEBUG
	printf("hashed: %d %d\n", hashed, hashed % HASH_TABLE_SIZE);
	#endif
	return hashed % HASH_TABLE_SIZE;

}


int lp_RemoveFrags(char *name)
/*	This function removes the "/xyz" part from the
	end of the name. */
{
	char	*s;
	
	s=strrchr(name,'/');
	if(s!=NULL) {
		*s='\0';
		return 1;
	} else
		return 0;
               
/*	int	length;
	
	length = strlen(name);
	while (name[length] != '/')
		name[length--] = 0;

	name[length] = 0; */
	
}

int lp_GetKills(Player *p)
{
		int a=0, i;
		for (i=0;i<NUMBER_OF_WEAPONS;i++)
			a += p->kills[i];
		
		return a;

}

int lp_GetDeaths(Player *p)
{
		int a=p->suicides, i;
		for (i=0;i<NUMBER_OF_WEAPONS;i++)
			a += p->deaths[i];
		
		return a;

}

float lp_GetKPerc(Player *p, int weapon)
{
	int	kills;
	kills = lp_GetKills(p);
	return 100.0 * p->kills[weapon] / kills;
}

float lp_GetDPerc(Player *p, int weapon)
{
	int	deaths;
	deaths = lp_GetDeaths(p);
	return 100.0 * p->deaths[weapon] / deaths;
}

float lp_GetTotalTime()
{
	Playerlist	*pl;
	int			i;
	float			ttime=0;
	
	for (i=0;i<HASH_TABLE_SIZE; i++) {
		pl = plist[i];
		while (pl != NULL) {
			ttime += pl->p->time_on;
			pl = pl->next;
		}
	}
	return ttime;
	
}


int lp_GetTKills()
{
	Playerlist	*pl;
	int			i, kills=0;
	
	for (i=0;i<HASH_TABLE_SIZE; i++) {
		pl = plist[i];
		while (pl != NULL) {
			kills += lp_GetKills(pl->p);
			pl = pl->next;
		}
	}
	return kills;
}

int lp_GetTDeaths()
{
	Playerlist	*pl;
	int			i, deaths=0;
	
	for (i=0;i<HASH_TABLE_SIZE; i++) {
		pl = plist[i];
		while (pl != NULL) {
			deaths += lp_GetDeaths(pl->p);
			pl = pl->next;
		}
	}
	return deaths;
}

float lp_GetScore(Player *p)
{
	return (p->death_score < 1.0 ? p->kill_score : 1.0 * p->kill_score / p->death_score);
}

float lp_Qrank(Player *p)
{
	float		ranking;
	if ((lp_GetDeaths(p)+p->drops) !=0) {
		ranking = (1.0 *  (lp_GetKills(p) + p->returns + 2*p->captures)) / (1.0 * (lp_GetDeaths(p) + p->drops));
		ranking *= 100;
	} else
		ranking = 0.0;
	
	return ranking;
}

float lp_TeamQuota(int TeamScore[], int MyTeam)
{
	float	a;
	
	switch (MyTeam) {
	case NO_TEAM	:	return 1;
							break;
	case RED_TEAM	: 	if (TeamScore[RED_TEAM]==0)
								return 1;
							else {
								a = 1.0 * TeamScore[BLUE_TEAM] / TeamScore[RED_TEAM];
								a = a > TQ_UPPER ? TQ_UPPER : a;
								a = a < TQ_LOWER ? TQ_LOWER : a;
								return (a == 0 ? 1.0 / TeamScore[RED_TEAM] : a);
							}
							break;
	case BLUE_TEAM	:	if (TeamScore[BLUE_TEAM]==0)
								return 1;
							else {
								a =  1.0 * TeamScore[RED_TEAM] / TeamScore[BLUE_TEAM];
								a = a > TQ_UPPER ? TQ_UPPER : a;
								a = a < TQ_LOWER ? TQ_LOWER : a;
								return (a == 0 ? 1.0 / TeamScore[BLUE_TEAM] : a);
							}
							break;
	default			:	return 1;
							break;	
	}
}

float lp_KillQuota(Player *p1, Player *p2)
{
		float a;
		a =  lp_GetScore(p1)==0 ? 1.0 : 1.0 * lp_GetScore(p2) / lp_GetScore(p1);
/* DBG */
		#ifdef QUOTA_DEBUG
		printf("%s:%f:%f\n",p1->name,p1->kill_score,p1->death_score);
		printf("%s:%f:%f\n\n",p2->name,p2->kill_score,p2->death_score);
		#endif
		a = a > KQ_UPPER ? KQ_UPPER : a;
		a = a < KQ_LOWER ? KQ_LOWER : a;
		return (a == 0 ? 1 : a);
}

void lp_PrintEntirePlayer(Player *p)
{
	printf("Name: %15s      kills   |   deaths\n", p->name);
	printf("Axe             :         %4d     |  %4d\n", p->kills[AXE], p->deaths[AXE]);
	printf("Boomstick       :         %4d     |  %4d\n", p->kills[BOOMSTICK], p->deaths[BOOMSTICK]);
	printf("Super shotgun   :         %4d     |  %4d\n", p->kills[SUPER_SHOTGUN], p->deaths[SUPER_SHOTGUN]);
	printf("Nailgun         :         %4d     |  %4d\n", p->kills[NAILGUN], p->deaths[NAILGUN]);
	printf("Super Nailgun   :         %4d     |  %4d\n", p->kills[SUPER_NAILGUN], p->deaths[SUPER_NAILGUN]);
	printf("Grenade Launcher:         %4d     |  %4d\n", p->kills[GRENADE_LAUNCHER], p->deaths[GRENADE_LAUNCHER]);
	printf("Rocket Launcher :         %4d     |  %4d\n", p->kills[ROCKET_LAUNCHER], p->deaths[ROCKET_LAUNCHER]);
	printf("Lightningbolt   :         %4d     |  %4d\n", p->kills[THUNDERBOLT], p->deaths[THUNDERBOLT]);
	printf("Grappling Hook  :         %4d     |  %4d\n", p->kills[GRAPPLING_HOOK], p->deaths[GRAPPLING_HOOK]);
	printf("Telefrags       :         %4d     |  %4d\n", p->kills[TELEFRAG], p->deaths[TELEFRAG]);
	printf("Killscore: %f      Deathscore: %f\n", p->kill_score, p->death_score);
	printf("Pickups: %d   Drops: %d   Recoveries: %d   Captures: %d\n",p->pickups,p->drops,p->returns,p->captures);
	printf("Suicides: %d\n", p->suicides);
}

void lp_ParseLogFile()
/* The main function for parsing all the entries in the
	current logfile. */
{

int		TeamScore[3];						/* 0 = noteam */ 
char		*str, *freestr;
FILE		*logf;
char		killer[256], killee[256], weapon[256];
Player	*Killer, *Killee;
float		killer_kill_score, ontime;


/*	Initialize parse string and open logfile */

	if ((str=malloc(512))==NULL)
		io_Error("Couldn\'t allocate memory for parse string.\n");
	if ((logf=fopen(files.logfile, "r"))==NULL)
		io_Error("Couldn\'t open logfile for reading.\n");

	freestr = str; /* So we can easily free the memory later */
	TeamScore[NO_TEAM]=0; /* No score for the teams yet */
	TeamScore[RED_TEAM]=globs.red_score;
	TeamScore[BLUE_TEAM]=globs.blue_score;

/*	Here we do all the parsing. Hold on to your hat. */
	while (!io_GetNextEntryLine(logf, str))
	{
	/* First reset the temporary Player objects. */
	
		Killer = NULL;
		Killee = NULL;

		if (sscanf(str, "LOG:\tDEATH\t%[^\t]\t%[^\t]\t%[^\n]\n", killee, killer, weapon)==3)
		{
			if (lp_RemoveFrags(killer))
			if (lp_RemoveFrags(killee))
			{	
			#ifdef DEBUG
			lp_MatchName(killer);
			lp_MatchName(killee);
			printf("%s was killed by %s in the library using the %s\n", killee, killer, weapon);
			#endif
			
			Killer = db_FindElseAddPlayer(killer);
			Killee = db_FindElseAddPlayer(killee);
			
			if (strcmp(weapon, S_AXE)==0) {
				Killer->kills[AXE]++;
				Killee->deaths[AXE]++;
			} else
			if (strcmp(weapon, S_BOOMSTICK)==0) {
				Killer->kills[BOOMSTICK]++;
				Killee->deaths[BOOMSTICK]++;
			} else
			if (strcmp(weapon, S_SUPER_SHOTGUN)==0) {
				Killer->kills[SUPER_SHOTGUN]++;
				Killee->deaths[SUPER_SHOTGUN]++;
			} else
			if (strcmp(weapon, S_NAILGUN)==0) {
				Killer->kills[NAILGUN]++;
				Killee->deaths[NAILGUN]++;
			} else
			if (strcmp(weapon, S_SUPER_NAILGUN)==0) {
				Killer->kills[SUPER_NAILGUN]++;
				Killee->deaths[SUPER_NAILGUN]++;
			} else
			if (strcmp(weapon, S_GRENADE_LAUNCHER)==0) {
				Killer->kills[GRENADE_LAUNCHER]++;
				Killee->deaths[GRENADE_LAUNCHER]++;
			} else
			if (strcmp(weapon, S_ROCKET_LAUNCHER)==0) {
				Killer->kills[ROCKET_LAUNCHER]++;
				Killee->deaths[ROCKET_LAUNCHER]++;
			} else
			if (strcmp(weapon, S_THUNDERBOLT)==0) {
				Killer->kills[THUNDERBOLT]++;
				Killee->deaths[THUNDERBOLT]++;
			} else
			if (strcmp(weapon, S_GRAPPLING_HOOK)==0) {
				Killer->kills[GRAPPLING_HOOK]++;
				Killee->deaths[GRAPPLING_HOOK]++;
			} else
			if (strcmp(weapon, S_TELEFRAG)==0) {
				Killer->kills[TELEFRAG]++;
				Killee->deaths[TELEFRAG]++;
			}


/*			Calculate the score for fragging someone. Make sure no one gets
			more than Q_UPPER pts and less than Q_LOWER;			
*/
			killer_kill_score = lp_TeamQuota(TeamScore, Killer->current_team)
										* lp_KillQuota(Killer, Killee);
										
			killer_kill_score = killer_kill_score > Q_UPPER ? Q_UPPER : killer_kill_score;
			killer_kill_score = killer_kill_score < Q_LOWER ? Q_LOWER : killer_kill_score;
			Killer->kill_score += killer_kill_score;

			Killee->death_score += killer_kill_score;

/*			Embarrasing bug.. Removed :-) */

/*			killee_death_score = killee_death_score > Q_UPPER ? Q_UPPER : killee_death_score;
			killee_death_score = killee_death_score < Q_LOWER ? Q_LOWER : killee_death_score;
			Killee->death_score += killee_death_score;*/
			
			TeamScore[Killer->current_team]++;
			
/* DBG */
			#ifdef QUOTA_DEBUG
			printf("KQ: %f %f\n", lp_KillQuota(Killer, Killee), lp_KillQuota(Killee, Killer));
			#endif
			}
		} else
		if (sscanf(str, "LOG:\tDEATH\t%[^\t]\t%[^\n]\n", killer, weapon)==2)
		{
			if (lp_RemoveFrags(killer))
			{
				#ifdef DEBUG
				lp_MatchName(killer); 
				printf("%s went japaneese with the %s\n", killer, weapon);
				#endif
			
				Killer = db_FindElseAddPlayer(killer);
			
				if ( strcmp(weapon, "teamchange") == 0 ) {
					if (Killer->current_team == BLUE_TEAM)
					{
						if (TeamScore[BLUE_TEAM] < TeamScore[RED_TEAM])
							Killer->chickens++;
						else
							Killer->braves++;
					} else
					if (Killer->current_team == RED_TEAM)
					{
						if (TeamScore[RED_TEAM] < TeamScore[BLUE_TEAM])
							Killer->chickens++;
						else
							Killer->braves++;
					}		
				} else {
					Killer->suicides++;
					Killer->death_score += 1.0 / lp_TeamQuota(TeamScore, Killer->current_team);
				}
			}
		} else
		if (sscanf(str, "LOG:\tCONNECT\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s just got connected\n", killer);
			#endif

		} else
		if (sscanf(str, "LOG:\tDISCONNECT\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s was just banished to the real world\n", killer);
			#endif
			Killer = db_FindElseAddPlayer(killer);
			Killer->current_team = NO_TEAM;
			Killer->connects++;
			/* There are connects after every map. Disconnects are more accurate. */
		} else
		if ( (sscanf(str, "FindFile: ./%[^/]/progs.%s", killer, killee) == 2) ||
			(sscanf(str, "FindFile: ./%[^/]/qwprogs.%s", killer, killee) == 2) )
		{
			/* Check for a map change.. Very ugly way.. */
			TeamScore[NO_TEAM]=0; /* No score for the teams on this map */
			TeamScore[RED_TEAM]=0;
			TeamScore[BLUE_TEAM]=0;
			globs.flagreg = 0; /* New level */
		} else
		if (sscanf(str, "LOG:\tFLAG-PICKUP\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s picked up the flag\n", killer);
			#endif
			
			Killer = db_FindElseAddPlayer(killer);
			Killer->pickups++;
		} else
		if (sscanf(str, "LOG:\tFLAG-DROP\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s dropped the flag\n", killer);
			#endif
			
			Killer = db_FindElseAddPlayer(killer);
			Killer->drops++;
			Killer->death_score += 1.0 / (FLAG_DROP * lp_TeamQuota(TeamScore, Killer->current_team));
			TeamScore[Killer->current_team==RED_TEAM ? BLUE_TEAM : RED_TEAM] +=TEAM_FLAG_DROP;
		} else
		if (sscanf(str, "LOG:\tFLAG-RECOVERY\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s recovered the flag\n", killer);
			#endif
			
			Killer = db_FindElseAddPlayer(killer);
			Killer->returns++;
			Killer->kill_score += FLAG_RECOVERY * lp_TeamQuota(TeamScore, Killer->current_team);
			TeamScore[Killer->current_team] += TEAM_FLAG_RECOVERY;
		} else
		if (sscanf(str, "LOG:\tFLAG-CAPTURE\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s captured the flag\n", killer);
			#endif
			
			Killer = db_FindElseAddPlayer(killer);
			Killer->captures++;
			Killer->kill_score += FLAG_CAPTURE * lp_TeamQuota(TeamScore, Killer->current_team);
			TeamScore[Killer->current_team] += TEAM_FLAG_CAPTURE;
		} else
		if (sscanf(str, "LOG:\tRED\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s changed to RED team", killer);
			#endif
			
			Killer = db_FindElseAddPlayer(killer);

			Killer->current_team=RED_TEAM;
		} else
		if (sscanf(str, "LOG:\tBLUE\t%[^\n]\n", killer)==1)
		{
			#ifdef DEBUG
			lp_MatchName(killer);
			printf("%s changed to BLUE team", killer);
			#endif

			Killer = db_FindElseAddPlayer(killer);

			Killer->current_team=BLUE_TEAM;
		} else
		if (sscanf(str, "RED team %s", killer) == 1)
		{
			if (strcmp(killer, "won")==0)
				globs.flagreg = 3;
			else
				globs.flagreg = 5;
		   /* The level is ending */
		} else
		if (sscanf(str, "O team %s", killer) == 1)
		{
			if (strcmp(killer, "won")==0)
				globs.flagreg = 3;
			else
				globs.flagreg = 5;
			/*	The level is ending */
		} else
		if (sscanf(str, "AIO and O team %s", killer) == 1)
		{
			globs.flagreg = 7; /* Teams are tied */
		} else
		if (sscanf(str, "LOG:\tTIME\t%f\t%[^\n]\n", &ontime, killer) == 2)
		{
			Killer = db_FindElseAddPlayer(killer);
			Killer->time_on += ontime;

			/* End of level.. Check if the player won or lost. */
			if ((Killer->current_team != NO_TEAM) && (globs.flagreg & 1) )
			{
/* Some debug crap.. It worked anyway :) */
/*				printf("flagreg=%d\n", globs.flagreg);
				printf("Before: current_team = %d\n", Killer->current_team);
				printf("victories = %d, defeats = %d, ties = %d\n",Killer->victories, Killer->defeats, Killer->ties);
*/				
				if (Killer->current_team == RED_TEAM)
					{
						Killer->victories += (((globs.flagreg & 7)==3) ? 1 : 0);
						Killer->defeats += (((globs.flagreg & 7)==5) ? 1 : 0);
						Killer->ties += (((globs.flagreg & 7)==7) ? 1 : 0);
					} else
					{
						Killer->victories += (((globs.flagreg & 7)==5) ? 1 : 0);
						Killer->defeats += (((globs.flagreg & 7)==3) ? 1 : 0);
						Killer->ties += (((globs.flagreg & 7)==7) ? 1 : 0);
					}
/*				printf("After: current_team = %d\n", Killer->current_team);
				printf("victories = %d, defeats = %d, ties = %d\n",Killer->victories, Killer->defeats, Killer->ties);
*/				
			}
			
		}
	}

	globs.red_score = TeamScore[RED_TEAM];
	globs.blue_score = TeamScore[BLUE_TEAM];

	
/* Cleanup routines */
	free(freestr);
	fclose(logf);
}

void lp_FixunixdedName(char *name)
/*	Unixded converts "dangerous" chars to [ASCII#]. We have to
	convert them back. */
{	
	char	work[NAME_LENGTH];
	int	i=0, w=0;
	unsigned char	asciin;
	
/*	printf("Before: %s ", name);fflush(stdout);*/
	
	while ( name[i] != 0 ) {
		if ( name[i] == '[' ) {
			if ( (name[++i] >= '0') && (name[i] <= '9') ) {
				asciin = strtol(&name[i], (char **)NULL, 16);
				while (name[i++] != ']');
				if (asciin)
					work[w++]=asciin;
				else
					work[w++]='[';
			} else {
				work[w++]=name[--i];
				i++;
			}
		} else
			work[w++]=name[i++];
	}
	work[w]=0;
	strcpy(name, work);
	
/*	printf("After: %s", name); */
}

void lp_MatchName(char *name)
{
/*	Removes illegal characters and "de-quakifies" special characters. */

	     		 /*0         1         2         3          4         5         6         7         8         9         10        11        12        13        14        15        16        17        18        19        20        21        22        23        24        25      */
             /*0123456789012345678901234567890123 4567890123456789012345678901234567890123456789012345678901 2345678901234567890123456789012345678901234567890123456789012345678901 2345678901234567890123456789012345678901234567890123456789 0123456789012345678901234567890123456789*/
char *convstr="-*]**.****.***..[]0123456789.[=] !\"#$%*'()*+,-./0123456789:;[=]?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~*[=]**.**** ***..[]0123456789.[=] !\"#$%*'()*+,-./0123456789:;[=]?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~*";

	int i,c;
	for(i=0;i<strlen(name);i++) {
		c = name[i];
		if(c<0)
			c=c+256;
		name[i] = convstr[c];
	}
	while (name[--i]==0x20) /* Remove trailing space */
		name[i]=0;

/*	Embarrasing bugfix for 1.15 */	
	if ( strcmp(name, "") == 0)
		strcpy(name, "(null)");
	
}


void lp_RetrieveStats()
{
	int					i;

/* NULLify the Hashtable. */
	for (i=0;i<HASH_TABLE_SIZE;i++)
		plist[i]=NULL;

 	/* Check if the logfile exists before opening it.
	   If the database file exists. Parse it first, else create it.
	   Just overwrite the html-file if that exists. 
	*/

	if (files.dbfile!=NULL){	/* Only enter database mode if a file is specified. */
		globs.dbused=1;
		if (io_FileExists(files.dbfile)) {
			#ifdef VERBOSE
			printf("[Reading database %s...]\n", files.dbfile);
			#endif
			db_ReadDBase(); 
		}
	}

/* Now read all entries in the current logfile,
	putting them in the list of Players.
	This function returns the total number of players. */

	if (files.logfile != NULL) {
		#ifdef VERBOSE
		printf("[Parsing logfile %s...]\n", files.logfile);
		#endif
		lp_ParseLogFile(plist);
	}

/*	We always have a html-file to write to. */

	#ifdef VERBOSE
	printf("[Writing html-header to %s...]\n", files.htmlfile);
	#endif
	html_WriteHeader(); 

	table = calloc(globs.nbrplayers, sizeof(Player *));	
	#ifdef VERBOSE
	printf("[Quicksorting player table...]\n");
	#endif

/*	Return how many entities are left after filtering away minkills. */

	globs.ents_left = db_MakeTable();

/*	Get the total number of deaths, kills and combined playing time. */
		
	globs.tkills  = lp_GetTKills();
	globs.tdeaths = lp_GetTDeaths(plist);
	globs.ttime = lp_GetTotalTime(plist);
		
/*	Finally write ourselves a html-table */
	#ifdef VERBOSE
	printf("[Writing table to html-file %s...]\n", files.htmlfile);
	#endif
	html_WriteTable();
				
/* If we are going to write a compact database,
	truncate the logfile so it can't be read again. */
	
	if (globs.dbused){
		if (inits.truncate)
			io_TruncateLogFile(files.logfile);
		#ifdef VERBOSE
		printf("[Writing Database to %s...]\n",files.dbfile);
		#endif VERBOSE
		db_WriteDBase(plist);
	} 
/* Now free all memory allocated */
	#ifdef VERBOSE
	printf("[Freeing resources and shutting down...]\n");
	#endif
	db_Free(plist);
	free(table);
}