#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cr.h"
#include "logparse.h"
#include "io.h"
#include "field.h"
#include "global.h"

/* Globals */
struct	file_s	files = {NULL, "ctfrank.html", NULL, NULL, "Quake", NULL, NULL};
struct	inits_s	inits = {50, 200, SO_BEST, SI_QRANK, 0};
struct	globs_s	globs = {0, 0, 0, {0,}, 0, 0.0, 0, 0, 0, 0};
Playerlist	*plist[HASH_TABLE_SIZE]; 
Player		**table=NULL;

void DisplayHelp(void)
{
	printf("CTFrank Parser v%.2f by Magnus Landqvist\n\n", CR_VERSION);
	printf("Usage: ctfrank [-?] [-order sort-order] [-item sort-item] <-dbdatafile.ext>\n");
	printf("       [-fields html-fields+..] [-minkills minkills] [-html outlist.html]\n");	
	printf("       [-maxents maxentries] [-cgi cgi-url] [-cgiconf identifier]\n");
	printf("       [-header header.html] [-foot foot.hml] [config ctfrank.cfg] [-trunc]\n");
	printf("       <-log logfile.ext>\n\n");
	printf("       -? displays this screen\n");
	printf("       -order sort order: <best|worst>\n");
	printf("       -item sort item :  <frank|kills|deaths|suicides|captures|drops\n"); 
	printf("          |pickups|recoveries|kill_axe|kill_boomstick|kill_super_shotgun\n");
	printf("          |kill_nailgun|kill_super_nailgun|kill_grenade|kill_rocket\n");
	printf("          |kill_bbq|kill_hook|kill_telefrag|death_(same as kills)\n");
	printf("          |killp_(same as kils)|deathp_(same as kills)|connects\n");
	printf("          |time_on|qrank|kills/min|deaths/min|victories|defeats|ties\n");
	printf("          |chickens|braves|capratio>\n");
	printf("       -fields html-fields: Same as sort item, but you can add more than one.\n");
	printf("          Separate with \"+\".\n");
	printf("       -minkills minimum kills, >= 0. Default is 50.\n");
	printf("       -maxents maximum entries. Default is 200.\n");
	printf("       -trunc Truncate the logfile after parsing it. Useful for continous logs.\n");
	printf("       -db r/w database set to datafile.ext. Not necessary.\n");
	printf("       -cgi cgi-url is the url to the program that shows personal stats.\n");
	printf("       -cgiconf config-type for the cgi. Default is QUAKE.\n");
	printf("       -html writes computed stats to outlist.html. Default is ctfrank.html\n");
	printf("       -header additional header for the html-file.\n");
	printf("       -foot additional foot for the html-file.\n");
	printf("       -log logfile.ext is the Threewave CTF logfile to parse.\n");
	printf("       -config use parameters from specified config file\n\n");
	exit(1);
}


int main(int argc, char *argv[])
{

	
	/* Clear those bloody fields, and allocate some mem */
/*	if ((fields = calloc(SI_FIELD_LENGTH, sizeof(int *))) == NULL)
		io_Error("Line 5?? You really should buy some more memory.\n"); */
		
	/* Set the standard fields */
	
	GetFields("kills+deaths+captures+drops+recoveries+frank");

	if (argc==1)
		DisplayHelp();
	else
	for (argv++,argc--; argc && **argv == '-'; argv++,argc--)
		if( (*argv)[1] == '?' )
			DisplayHelp();
		else
		if ( (io_stricmp(*argv, "-html") == 0) && argc-- > 1 )
			files.htmlfile = *++argv;
		else
		if ( (io_stricmp(*argv, "-db") == 0) && argc-- > 1 )
			files.dbfile = *++argv;
		else
		if ( (io_stricmp(*argv, "-log") == 0) && argc-- > 1 )
			files.logfile = *++argv;
		else
		if ( (io_stricmp(*argv, "-cgiconf") == 0) && argc-- > 1 )
			files.cgiconf = *++argv;
		else
		if ( (io_stricmp(*argv, "-header") == 0) && argc-- > 1 )
			files.htmlheader = *++argv;
		else
		if ( (io_stricmp(*argv, "-foot") == 0) && argc-- > 1 )
			files.htmlfoot = *++argv;
		else
		if ( (io_stricmp(*argv, "-order") == 0) && argc-- > 1 )
			inits.sortorder = GetSO(*++argv);
		else
		if ( (io_stricmp(*argv, "-item") == 0) && argc-- > 1 )
			inits.sortitem = GetSI(*++argv);
		else
		if ( (io_stricmp(*argv, "-minkills") == 0) && argc-- > 1 )
			inits.minkills = atoi(*++argv);
		else
		if ( (io_stricmp(*argv, "-fields") == 0) && argc-- > 1 )
			GetFields(*++argv);
		else
		if ( (io_stricmp(*argv, "-maxents") == 0) && argc-- > 1 )
			inits.maxentries = atoi(*++argv);
		else
		if ( (io_stricmp(*argv, "-cgi") == 0) && argc-- > 1 )
			files.cgi_url = *++argv;
		else
		if ( (io_stricmp(*argv, "-trunc") == 0) )
			inits.truncate = 1;
		else
		if ( (io_stricmp(*argv, "-config") == 0) && argc-- > 1 )
			GetConfig(*++argv);
		else
			io_Error("Invalid argument or missing parameter.\n");

	if (files.logfile == NULL && files.dbfile == NULL)
		io_Error("You must specify a log file and/or a database file.\n");
	
	lp_RetrieveStats();

	return 0;
}
